#include <inttypes.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>

static void skipspace(FILE *f)
{
	int c;
	while ( (c = fgetc(f)) != EOF ) {
		if (!isspace(c)) {
			ungetc(c, f);
			return;
		}
	}
}

static int getwidth(const char **pfmt)
{
	int w;

	for (w = 0; isdigit(**pfmt); (*pfmt)++)
		w = w * 10 + (**pfmt - '0');
	return w;
}

static int getaccept(const char **pfmt, char accept[])
{
	int neg = 0;
	int lastc = 0;
	int i;
	const char *fmt = *pfmt;

	for (i = 0; i < 256; i++)
		accept[i] = 0;

	if (*fmt == '^') {
		neg = 1;
		if (*++fmt == '\0')
			return -1;
	}
	if (*fmt == ']') {
		accept[(int)']'] = 1;
		if (*++fmt == '\0')
			return -1;
	}
	while (*fmt != '\0' && *fmt != ']') {
		if (*fmt != '-' || *(fmt + 1) == ']') {
			lastc = *fmt++;
			accept[lastc] = 1;
		} else {
			if (*++fmt == '\0')
				return -1;
			if (lastc > *fmt)
				return -1;
			for (lastc++; lastc <= *fmt; lastc++)
				accept[lastc] = 1;
			fmt++;
		}
	}
	if (*fmt != ']')
		return -1;
	if (neg) {
		for (i = 0; i < 256; i++)
			accept[i] = 1 - accept[i];
	}
	*pfmt = fmt;
	return 0;
}

static int parseuint(FILE *f, unsigned long long *p, int width, int base)
{
	int c, d, i;
	unsigned long long v;

	if (width < 0)
		width = INT_MAX;
	
	v = 0;
	for (i = 0; i < width && (c = fgetc(f)) != EOF; i++) {
		d = -1;
		if (isdigit(c)) {
			d = c - '0';
		} else if (isalpha(c)) {
			d = tolower(c) - 'a';
		}
		if (d < 0 || d >= base) {
			ungetc(c, f);
			break;
		}
		v = v * base + d;
	}
	if (i == 0)
		return -1;
	*p = v;
	return 0;
}

static int parseint(FILE *f, long long *p, int width, int base)
{
	unsigned long long v;
	int c, neg;

	c = fgetc(f);
	neg = 0;
	if (c == '-') {
		neg = 1;
	} else if (c == EOF) {
		return -1;
	} else {
		ungetc(c, f);
	}
	if (parseuint(f, &v, width, base) < 0)
		return -1;
	*p = neg ? -(long long)v : (long long)v;
	return 0;
}

int fscanf(FILE *f, const char *fmt, ...)
{
	int count, n;
	va_list ap;
	int c;
	int noassign;
	int width;
	int modifier;
	int doubled;
	void *output;
	long long iscalar;
	unsigned long long uscalar;
	int base;
	char accept[256];

	count = 0;
	va_start(ap, fmt);
	c = 0;
	while (c != EOF && *fmt != '\0') {
		if (isspace(*fmt)) {
			fmt++;
			skipspace(f);
			continue;
		}
		if (*fmt != '%') {
			c = fgetc(f);
			if (c != *fmt)
				goto end;
			fmt++;
			continue;
		}
		fmt++;
		noassign = 0;
		if (*fmt == '*') {
			if (*++fmt == '\0')
				goto error;
			noassign = 1;
		}
		// maximum field width
		width = -1;
		if (isdigit(*fmt)) {
			width = getwidth(&fmt);
			if (*fmt == '\0')
				goto error;
		}
		// type modifier
		modifier = 0;
		doubled = 0;
		if (strchr("hjlLtz", *fmt)) {
			modifier = *fmt++;
			if ((modifier == 'h' || modifier == 'l') && *fmt == modifier) {
				doubled = 1;
				fmt++;
			} else if (modifier == 'L') {
				modifier = 'l';
				doubled = 1;
			}
			if (*fmt == '\0')
				goto error;
		}
		// get the pointer if necessary
		output = NULL;
		if (!noassign)
			output = va_arg(ap, void *);
		// parse input
		switch (*fmt) {
		case '%':
			skipspace(f);
			if (fgetc(f) != '%')
				goto end;
			break;
		case 'd':
			skipspace(f);
			if (parseint(f, &iscalar, width, 10) < 0)
				goto end;
			break;
		case 'i':
			skipspace(f);
			base = 10;
			if ( (c = fgetc(f)) == '0') {
				if ( (c = fgetc(f)) == 'x' || c == 'X' ) {
					base = 16;
				} else {
					ungetc(c, f);
					base = 8;
				}
			} else {
				ungetc(c, f);
			}
			if (parseint(f, &iscalar, width, base) < 0)
				goto end;
			break;
		case 'o':
			skipspace(f);
			if (parseuint(f, &uscalar, width, 8) < 0)
				goto end;
			break;
		case 'u':
			skipspace(f);
			if (parseuint(f, &uscalar, width, 10) < 0)
				goto end;
			break;
		case 'x':
		case 'X':
			skipspace(f);
			if ( (c = fgetc(f)) != '0' ) {
				ungetc(c, f);
			} else if ( (c = fgetc(f)) != 'x' && c != 'X' ) {
				ungetc(c, f);
			}
			if (parseuint(f, &uscalar, width, 16) < 0)
				goto end;
			break;
		case 's':
			skipspace(f);
			if (width < 0)
				width = INT_MAX;
			for (n = 0; n < width && (c = fgetc(f)) != EOF; n++) {
				if (isspace(c)) {
					ungetc(c, f);
					break;
				}
				if (output != NULL) {
					*((char *)output) = c;
					output = (char *)output + 1;
				}
			}
			if (n > 0) {
				if (output != NULL)
					*(char *)output = '\0';
				count++;
			}
			break;
		case 'c':
			if (width < 0)
				width = 1;
			for (n = 0; n < width && (c = fgetc(f)) != EOF; n++) {
				if (output != NULL) {
					*((char *)output) = c;
					output = (char *)output + 1;
				}
			}
			if (n > 0)
				count++;
			break;
		case '[':
			fmt++;
			if (getaccept(&fmt, accept) < 0)
				goto error;
			if (width < 0)
				width = INT_MAX;
			for (n = 0; n < width && (c = fgetc(f)) != EOF; n++) {
				if (!accept[c]) {
					ungetc(c, f);
					break;
				}
				if (output != NULL) {
					*((char *)output) = c;
					output = (char *)output + 1;
				}
			}
			if (n > 0) {
				if (output != NULL)
					*(char *)output = '\0';
				count++;
			}
			break;
		case 'p':
			if (fgetc(f) != '0' || ( (c = fgetc(f)) != 'x' && c != 'X'))
				goto end;
			if (parseuint(f, &uscalar, width, 16) < 0)
				goto end;
			*(void **)output = (void *)uscalar;
			count++;
			break;
		case 'n':
			break;
		default:
			goto error;
		}
		// assign scalar values
		if (output != NULL) {
			if (*fmt == 'd' || *fmt == 'i') {
				switch (modifier) {
				case 0:
					*(int *)output = iscalar;
					break;
				case 'h':
					if (doubled)
						*(short *)output = iscalar;
					else
						*(signed char *)output = iscalar;
					break;
				case 'j':
					*(intmax_t *)output = iscalar;
					break;
				case 'l':
					if (doubled)
						*(long long *)output = iscalar;
					else
						*(long *)output = iscalar;
					break;
				case 't':
					*(ptrdiff_t *)output = iscalar;
					break;
				case 'z':
					*(ssize_t *)output = iscalar;
					break;
				}
				count++;
			} else if (strchr("ouxX", *fmt) != NULL) {
				switch (modifier) {
				case 0:
					*(unsigned int *)output = uscalar;
					break;
				case 'h':
					if (doubled)
						*(unsigned short *)output = uscalar;
					else
						*(unsigned char *)output = uscalar;
					break;
				case 'j':
					*(uintmax_t *)output = uscalar;
					break;
				case 'l':
					if (doubled)
						*(unsigned long long *)output = uscalar;
					else
						*(unsigned long *)output = uscalar;
					break;
				case 't':
					*(uptrdiff_t *)output = uscalar;
					break;
				case 'z':
					*(size_t *)output = uscalar;
					break;
				}
				count++;
			}
		}
		fmt++;
	}
end:
	va_end(ap);
	return count ? count : EOF;
error:
	va_end(ap);
	return EOF;
}
