#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* parse permissions in symbolic form */

mode_t parse_perms(char *p, mode_t tm)
{
	mode_t cumask = 0;
	int is_dir = S_ISDIR(tm);
	tm = tm & 07777;

	while (*p != '\0') {
		int i, who, perm, sflag, tflag, Xflag;
		char op;

		/* first, a sequence of zero or more chars in [ugoa].  These
		 * determine which set(s) of permissions we need to change. We
		 * store this information in a 'who' bitmap: bit 2 for u, bit 1
		 * for g and bit 0 for o.
		 */
		who = 0;
		for (;;) {
			if (*p == 'u')
				who |= 4;
			else if (*p == 'g')
				who |= 2;
			else if (*p == 'o')
				who |= 1;
			else if (*p == 'a')
				who = 7;
			else
				break;
			p++;
		}
		if (!who) {
			/* by default apply to all, but use the current umask
			 * (Note that the umask is not used otherwise)
			 */
			who = 7;
			/* there is no way to read the current umask, other than
			 * setting it to something and get the previous one.
			 */
			cumask = umask(0);
			umask(cumask); // write it back
		}
		/* now a sequence of one or more actions, where each action is
		 * an operation (one of =, + or -) followed by a set of
		 * permissions (zero or more of r, w, x, X, s, t, u, g, o).  We
		 * remember the operation in 'op' and collect the permissions
		 * in the 'perm' bitmap. The 's', 't' and 'X' flags must be
		 * handled specially at a later stage.
		 */
		op = -1; // if 'op' is still -1 after this cycle, we
			 // didn't see any action
		while (*p != '\0' && strchr("=+-", *p)) {
			op = *p++;
			perm = 0;
			sflag = tflag = Xflag = 0;
			while (*p != '\0' && !strchr("=+-,", *p)) {
				switch (*p) {
				case 'r': perm |= 4; break;
				case 'w': perm |= 2; break;
				case 'x': perm |= 1; break;
				case 'X': Xflag = 1; break;
				case 's': sflag = 1; break;
				case 't': tflag = 1; break;
				case 'u': perm = (tm >> 6) & 07; break;
				case 'g': perm = (tm >> 3) & 07; break;
				case 'o': perm = (tm >> 0) & 07; break;
				default:
					fprintf(stderr, "syntax error\n");
					exit(1);
				}
				p++;
			}
			/* now apply 'perm' to 'who' according to the selected
			 * operation (as stored in 'op')
			 *
			 * The 'for' is on the bits of 'who'.
			 */
			for (i = 0; i < 3; i++) {
				mode_t pm, mask;

				if (!(who & (1 << i)))
					continue;
				mask = (7 << (i * 3)) | (1 << (i + 9));
				// if X was used, add 'x' to the set of permission
				// if the file is a directory or if any other 'x'
				// permission is also set
				if (Xflag && (is_dir || (tm & 0111)))
					perm |= 1;
				pm = perm << (i * 3);
				if (sflag) { // set-uid or set-gid flag
					if (i == 1) // user
						pm |= 02000;
					else if (i == 2) // group
						pm |= 04000;
					// ignored otherwise
				}
				if (tflag && i == 1) // sticky-bit
					pm |= 01000;
				switch (op) {
				case '=':
					tm &= ~mask;
					/* fallthrough */
				case '+':
					tm |= pm;
					break;
				case '-':
					tm &= ~pm;
					break;
				}
			}
		}
		if (op == -1) {
			fprintf(stderr, "syntax error\n");
			exit(1);
		}
		if (*p == ',') // another clause follows
			p++;
	}
	// return the computed new-mode, as modified by the current user
	// mask, if necessary (note that cumask is still zero if it is not
	// needed and '& ~cumask' is a NOP in that case).
	return tm & ~cumask;
}

int main(int argc, char* argv[])
{
	int i, needstat;
	mode_t tm = 0;
	struct stat statbuf;

	if (argc < 3) {
		fprintf(stderr, "usage: chmod <mode> <file>...\n");
		exit(1);
	}

	// to process symbolic input in general, we need to stat() every file
	// to get its current permissions.  We can avoid that if the input is
	// numeric. 
	needstat = 1;
	if (isdigit(argv[1][0])) {
		tm = strtol(argv[1], NULL, 8);
		needstat = 0;
	}

	for (i = 2; i < argc; i++) {
		if (needstat) {
			if (stat(argv[i], &statbuf) < 0) {
				perror(argv[i]);
				continue;
			}
			tm = parse_perms(argv[1], statbuf.st_mode);
		}
		if (chmod(argv[i], tm) < 0) 
			perror(argv[i]);
	}
	return 0;
}
