#define MAX_LINE 1024

struct input {
	// Each input object reads either from a file (fd) or from a buffer in
	// memory (mem)
	int fd;			// -1 for objects that read from mem
	const char *mem;	// NULL for objects that read from fd

	// the current char
	int ibuf;
	// last char peeked (0 if none)
	int nibuf;
	// non-zero if we met an error
	// if set, it stops in/out until cleared by newcmd()
	// it can be set by the client
	int error;
	// non-zero if we need to prompt the user
	int interactive;
	// non-zero if we have just started collecting a new command
	int newcmd;

	// storage for oput(); oscan is where we will write next
	char obuf[MAX_LINE], *oscan;
	// pointer to the first char of the last token (inside obuf)
	char *tokval;
};

// initialize the input reader from a file.  We start at the beginning of a
// line, but newcmd() has not yet been called (so the newcmd flag is zero).
// The input is interactive iff the file is a tty (this can be changed by
// overwriting s->interactive, if desired).
void iinit(struct input *s, int fd)
{
	s->fd = fd;
	s->mem = NULL;
	s->interactive = isatty(s->fd);

	s->error = s->newcmd = 0;
	s->ibuf = '\n';
	s->nibuf = 0;
	s->oscan = s->obuf;
	s->tokval = NULL;
}

// initialize the input reader from a buffer in memory. mem must point
// to a '\0' terminated string. For these objects the input is always
// non interactive.
void iinit_mem(struct input *s, const char *mem)
{
	s->fd = -1;
	s->mem = mem;
	s->interactive = 0;

	// the rest of the initialization is the same as in iinit
	s->error = s->newcmd = 0;
	s->ibuf = '\n';
	s->nibuf = 0;
	s->oscan = s->obuf;
	s->tokval = NULL;
}

// get into s->ibuf the next non-null byte from the input, prompting the user
// if needed.  Returns the new char, or EOF at end of input. It may set
// s->error.
int igetc(struct input *s);

// put back c into the input stream
void iungetc(struct input *s, char c);

// load the next character into s->ibuf, skipping unquoted '\'-newline sequences.
void inext(struct input *s, int quoted)
{
	if (s->error)
		return;

	// if the shell non-interactive (e.g., parsing a script) input stops at
	// EOF (i.e., further calls to inext do nothing); interactive shells,
	// instead, exit only if the EOF was received on an empty line;
	// otherwise, they continue as usual.
	if (!s->interactive && s->ibuf == EOF)
		return;

again:
	switch (igetc(s)) {
	case EOF:
		if (s->error) {
			fprintf(stderr, "I/O error\n");
			return;
		}
		break;
	case '\\':
		if (!quoted) {
			if (igetc(s) == '\n')
				goto again;
			else
				iungetc(s, '\\');
		}
		break;
	default:
		break;
	}
	// we have read at least one char, so we are no longer at the beginning
	// of a new command
	s->newcmd = 0;
}

// start collecting a new command. Return zero if we are at EOF
int nextcmd(struct input *s)
{
	if (s->interactive) {
		// forget about any previous EOF
		if (s->ibuf == EOF)
			s->ibuf = '\n';
	}
	s->newcmd = 1;
	// drop the old obuf contents
	s->oscan = s->obuf;
	s->tokval = NULL;
	// reset errors
	s->error = 0;
	// preload ibuf and prompt the user with $PS1 if necessary
	if (s->ibuf == '\n')
		inext(s, 0);
	return s->ibuf != EOF;
}

// close the reader. When interactive, the last thing we received was a Ctrl+D
// and not a newline: in this case we print a newline before exiting, to
// visually improve the output.
void iclose(struct input *s)
{
	if (s->interactive)
		fprintf(stderr, "\n");
	close(s->fd);
}

// remaining space in obuf
size_t ospace(struct input *s)
{
	return MAX_LINE - (s->oscan - s->obuf);
}

// push a char into the obuf, checking for overflows
void oput(struct input *s, unsigned char c)
{
	if (s->error)
		return;

	if (ospace(s) <= 0) {
		fprintf(stderr, "line too long\n");
		s->error = 1;
		return;
	}
	*s->oscan++ = c;
}

// start collecting a new token
void newtoken(struct input *s)
{
	if (s->error)
		return;

	s->tokval = s->oscan;
}

// return a pointer to the last token
char *lasttoken(struct input *s)
{
	if (s->error)
		return NULL;

	return s->tokval;
}

// internal functions and implementation of igetc and iungetc

// return the next non-null byte from either fd or mem.
// Returns 0 at end of input and -1 in case of error.
int iread(struct input *s)
{
	char c;

	if (s->fd >= 0) {
		int n;

		while ( (n = read(s->fd, &c, 1)) > 0 && !c)
			;
		if (n <= 0)
			return n;
	} else {
		c = *s->mem++;
		if (c == '\0')
			return 0;
	}
	return (unsigned char)c;
}

int igetc(struct input *s)
{
	int c;

	// first, return any char that was 'ungetted'
	if ( (c = s->nibuf) ) {
		s->ibuf = s->nibuf;
		s->nibuf = 0;
		return c;
	}

	// We need to prompt the user for new chars whenever the input
	// is interactive and the last char read was '\n'.  The prompt
	// changes if we have just started collecting a new command
	// (s->newcmd == 1) or we are in the middle of command. In the
	// former case, the prompt also changes if the current effective
	// user is root. Non-root prompts can be customized using the
	// PS1 and PS2 environment variables.
	if (s->interactive && s->ibuf == '\n') {
		const char *ps;

		if (s->newcmd) {
			if (!geteuid()) // root user
				ps = "# ";
			else if ( (ps = getenv("PS1")) == NULL)
				ps = "$ ";
		} else {
			if ( (ps = getenv("PS2")) == NULL)
				ps = "> ";
		}
		fprintf(stderr, "%s", ps);
	}
	switch ( (c = iread(s)) ) {
	case -1:
		s->error = 1;
		/* fallthrough */
	case 0:
		s->ibuf = EOF;
		break;
	default:
		s->ibuf = c;
		break;
	}
	return s->ibuf;
}

void iungetc(struct input *s, char c)
{
	s->nibuf = s->ibuf;
	s->ibuf = c;
}
