#include <sys/wait.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h> /* for perror() */
#include <string.h>
#include <unistd.h>

#include "input.h"

/* shell that understands "cd" */

#define QUOTE 0x80
#define MAX_ARGS 10

void space(struct input *s)
{
	while (s->ibuf == ' ' || s->ibuf == '\t')
		inext(s, 0);
}

void collect_sq(struct input *s)
{
	// skip the initial quote
	inext(s, 1);
	// scan the ibuf until we find the closing single quote,
	// copying and marking all enclosed characters as we go
	while (s->ibuf != '\'') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing quote\n");
			s->error = 1;
			return;
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 1);
	}
	// skip the final quote
	inext(s, 0);
}

void collect_dq(struct input *s)
{
	// skip the initial double quote
	inext(s, 0);
	// scan the ibuf until we find the closing double quote,
	// copying and marking enclosed characters as we go
	while (s->ibuf != '"') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing double quote\n");
			s->error = 1;
			return;
		}
		if (s->ibuf == '\\') {
			// a backslash within double quotes removes the
			// special meaning from some characters.
			inext(s, 1);
			if (s->ibuf != '$' && s->ibuf != '"' && s->ibuf != '\\')
				oput(s, '\\' | QUOTE);
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 0);
	}
	// skip the final double quote
	inext(s, 0);
}

int word_delim(int c)
{
	return     c == EOF || isspace(c)
		;
}

int word(struct input *s)
{
	newtoken(s);
	while (!word_delim(s->ibuf)) {
		switch (s->ibuf) {
		case '\'':
			collect_sq(s);
			break;
		case '"':
			collect_dq(s);
			break;
		case '\\':
			// skip the backslash
			inext(s, 1);
			// copy and quote the next char
			oput(s, s->ibuf | QUOTE);
			inext(s, 0);
			break;
		default:
			oput(s, s->ibuf);
			inext(s, 0);
			break;
		}
	}
	// terminate the word
	oput(s, '\0');
	return 1;
}

int token(struct input *s)
{
	space(s);
	switch (s->ibuf) {
	case EOF:
	case '\n':
		return '\n';
	default:
		return word(s);
	}
}

void remove_quotes(char *w)
{
	while (*w != '\0')
		*w++ &= ~QUOTE;
}

int main()
{
	struct input s;
	int t, c_argc;
	char *c_argv[MAX_ARGS + 1];
	char *tok;

	iinit(&s, 0);
	while (nextcmd(&s) > 0) {
		c_argc = 0;
		while ( (t = token(&s)) != '\n') {
			if (s.error)
				break;
			switch (t) {
			case 1:
				if (c_argc >= MAX_ARGS) {
					fprintf(stderr, "too many arguments\n");
					s.error = 1;
					break;
				}
				tok = lasttoken(&s);
				remove_quotes(tok);
				c_argv[c_argc++] = tok;
			}
		}
		if (s.error) {
			if (!s.interactive)
				exit(1);
			while (token(&s) != '\n')
					;
			continue;
		}
		c_argv[c_argc] = NULL;
		if (c_argc < 1)
			c_argv[0] = "";

		if (strcmp(c_argv[0], "cd") == 0) {
			if (c_argc != 2) {
				fprintf(stderr, "usage: cd <dir>\n");
				continue;
			}
			if (chdir(c_argv[1]) < 0)
				perror(c_argv[1]);
			continue;
		}

		if (fork()) {
			wait(0);
		} else {
			if (c_argc < 1)
				exit(0);
			execvp(c_argv[0], c_argv);
			perror(c_argv[0]);
			exit(1);
		}
	}
	iclose(&s);
	return 0;
}
