#include <sys/wait.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h> /* for perror() */
#include <string.h>
#include <unistd.h>

#include "input.h"

/* shell with I/O redirection */

#define QUOTE 0x80
#define MAX_ARGS 10

void space(struct input *s)
{
	while (s->ibuf == ' ' || s->ibuf == '\t')
		inext(s, 0);
}

void collect_sq(struct input *s)
{
	// skip the initial quote
	inext(s, 1);
	// scan the ibuf until we find the closing single quote,
	// copying and marking all enclosed characters as we go
	while (s->ibuf != '\'') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing quote\n");
			s->error = 1;
			return;
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 1);
	}
	// skip the final quote
	inext(s, 0);
}

void collect_dq(struct input *s)
{
	// skip the initial double quote
	inext(s, 0);
	// scan the ibuf until we find the closing double quote,
	// copying and marking enclosed characters as we go
	while (s->ibuf != '"') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing double quote\n");
			s->error = 1;
			return;
		}
		if (s->ibuf == '\\') {
			// a backslash within double quotes removes the
			// special meaning from some characters.
			inext(s, 1);
			if (s->ibuf != '$' && s->ibuf != '"' && s->ibuf != '\\')
				oput(s, '\\' | QUOTE);
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 0);
	}
	// skip the final double quote
	inext(s, 0);
}

int word_delim(int c)
{
	return     c == EOF || isspace(c)
		|| c == '<'  || c == '>'
		;
}

int word(struct input *s)
{
	newtoken(s);
	while (!word_delim(s->ibuf)) {
		switch (s->ibuf) {
		case '\'':
			collect_sq(s);
			break;
		case '"':
			collect_dq(s);
			break;
		case '\\':
			// skip the backslash
			inext(s, 1);
			// copy and quote the next char
			oput(s, s->ibuf | QUOTE);
			inext(s, 0);
			break;
		default:
			oput(s, s->ibuf);
			inext(s, 0);
			break;
		}
	}
	// terminate the word
	oput(s, '\0');
	return 1;
}

int redir(struct input *s)
{
	int t;

	t = s->ibuf;
	inext(s, 0);
	if (t == '>') {
		if (s->ibuf == '>') {
			inext(s, 0);
			t = 'A';
		}
	}
	return t;
}

int token(struct input *s)
{
	space(s);
	switch (s->ibuf) {
	case EOF:
	case '\n':
		return '\n';
	case '<': case '>':
		return redir(s);
	default:
		return word(s);
	}
}

void remove_quotes(char *w)
{
	while (*w != '\0')
		*w++ &= ~QUOTE;
}

/* get the file name after a redirection operator */
char *getfname(struct input *s)
{
	char *f;
	/* the file name must be a word */
	if (token(s) != 1) {
		fprintf(stderr, "missing file name\n");
		s->error = 1;
		return NULL;
	}
	f = lasttoken(s);
	remove_quotes(f);
	return f;
}

int main()
{
	struct input s;
	int t, c_argc;
	char *c_argv[MAX_ARGS + 1];
	char *tok;
	const char *ifile, *ofile;
	int oflags;

	iinit(&s, 0);
	while (nextcmd(&s) > 0) {
		c_argc = 0;
		ifile = ofile = NULL;
		oflags = O_TRUNC;
		while ( (t = token(&s)) != '\n') {
			if (s.error)
				break;
			switch (t) {
			case 1:
				if (c_argc >= MAX_ARGS) {
					fprintf(stderr, "too many arguments\n");
					s.error = 1;
					break;
				}
				tok = lasttoken(&s);
				remove_quotes(tok);
				c_argv[c_argc++] = tok;
				break;
			case '<':
				if (ifile != NULL) {
					fprintf(stderr, "multiple input redirect\n");
					s.error = 1;
					break;
				}
				ifile = getfname(&s);
				break;
			case 'A':
				oflags = O_APPEND;
				/* fallthrough */
			case '>':
				if (ofile != NULL) {
					fprintf(stderr, "multiple output redirect\n");
					s.error = 1;
					break;
				}
				ofile = getfname(&s);
				break;
			}
		}
		if (s.error) {
			if (!s.interactive)
				exit(1);
			while (token(&s) != '\n')
					;
			continue;
		}
		c_argv[c_argc] = NULL;
		if (c_argc < 1)
			c_argv[0] = "";

		if (strcmp(c_argv[0], "cd") == 0) {
			if (c_argc != 2) {
				fprintf(stderr, "usage: cd <dir>\n");
				continue;
			}
			if (chdir(c_argv[1]) < 0)
				perror(c_argv[1]);
			continue;
		}

		if (fork()) {
			wait(0);
		} else {
			if (ifile != NULL) {
				close(0);
				if (open(ifile, O_RDONLY) < 0) {
					perror(ifile);
					exit(1);
				}
			}
			if (ofile != NULL) {
				close(1);
				if (open(ofile, O_WRONLY | O_CREAT | oflags, 0666) < 0) {
					perror(ofile);
					exit(1);
				}
			}
			if (c_argc < 1)
				exit(0);
			execvp(c_argv[0], c_argv);
			perror(c_argv[0]);
			exit(1);
		}
	}
	iclose(&s);
	return 0;
}
