#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h> /* for perror() */
#include <string.h>
#define _GNU_SOURCE
#include <unistd.h>

#include "input.h"

/* shell that handles the return value from commands */


#define QUOTE 0x80
#define MAX_ARGS 10
#define MAX_VARS 10

pid_t shellpid;
int last_result;

void varname(struct input *s)
{
	if (isalpha(s->ibuf) || s->ibuf == '_') {
		do {
			oput(s, s->ibuf);
			inext(s, 0);
		} while (isalnum(s->ibuf) || s->ibuf == '_');
	} else {
		if (s->ibuf != '\n' && s->ibuf != EOF) {
			oput(s, s->ibuf);
			inext(s, 0);
		}
	}
	// terminate the varname
	oput(s, '\0');
}

void space(struct input *s)
{
	while (s->ibuf == ' ' || s->ibuf == '\t')
		inext(s, 0);
}

void collect_sq(struct input *s)
{
	// skip the initial quote
	inext(s, 1);
	// scan the ibuf until we find the closing single quote,
	// copying and marking all enclosed characters as we go
	while (s->ibuf != '\'') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing quote\n");
			s->error = 1;
			return;
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 1);
	}
	// skip the final quote
	inext(s, 0);
}

void collect_dq(struct input *s)
{
	// skip the initial double quote
	inext(s, 0);
	// scan the ibuf until we find the closing double quote,
	// copying and marking enclosed characters as we go
	while (s->ibuf != '"') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing double quote\n");
			s->error = 1;
			return;
		}
		if (s->ibuf == '$') {
			inext(s, 0);
			if (s->ibuf == '"' || s->ibuf == EOF)
				continue;
			// ugly trick. The '$' may behave in three ways:
			//
			//   1) unquoted: expand the variable without
			//   	         quoting the resulting spaces
			//   2) quoted in dq: expand and quote everything
			//   3) quoted (sq or bs): do nothing
			//
			// but, we only have one spare bit. We use the QUOTE
			// bit to distinguish 1) and 3). For 2) we replace
			// the '$' with '"'. This works, because unQUOTEd
			// double quotes cannot otherwise appear in the obuf.
			oput(s, '"');
			varname(s);
			continue;
		}
		if (s->ibuf == '\\') {
			// a backslash within double quotes removes the
			// special meaning from some characters.
			inext(s, 1);
			if (s->ibuf != '$' && s->ibuf != '"' && s->ibuf != '\\')
				oput(s, '\\' | QUOTE);
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 0);
	}
	// skip the final double quote
	inext(s, 0);
}

int word_delim(int c)
{
	return     c == EOF || isspace(c)
		|| c == '<'  || c == '>'
		;
}

int word(struct input *s)
{
	newtoken(s);
	while (!word_delim(s->ibuf)) {
		switch (s->ibuf) {
		case '\'':
			collect_sq(s);
			break;
		case '"':
			collect_dq(s);
			break;
		case '$':
			oput(s, '$');
			inext(s, 0);
			varname(s);
			break;
		case '\\':
			// skip the backslash
			inext(s, 1);
			// copy and quote the next char
			oput(s, s->ibuf | QUOTE);
			inext(s, 0);
			break;
		default:
			oput(s, s->ibuf);
			inext(s, 0);
			break;
		}
	}
	// terminate the word
	oput(s, '\0');
	return 1;
}

int redir(struct input *s)
{
	int t;

	t = s->ibuf;
	inext(s, 0);
	if (t == '>') {
		if (s->ibuf == '>') {
			inext(s, 0);
			t = 'A';
		}
	}
	return t;
}

int token(struct input *s)
{
	space(s);
	switch (s->ibuf) {
	case EOF:
	case '\n':
		return '\n';
	case '<': case '>':
		return redir(s);
	default:
		return word(s);
	}
}

// expand the variable pointed to by *w, advancing the *w pointer
// past the terminating '\0' character. Return the variable
// value, or NULL if the variable is not set.
char *getvar(char **w)
{
	static char vbuf[MAX_LINE];
	char *e;

	// skip past the varname delimiter
	(*w)++;
	// get the value of the variable
	if (isalpha(**w))
		e = getenv(*w);
	else {
		// special variables
		e = vbuf;
		switch (**w) {
		case '$':
			snprintf(vbuf, MAX_LINE, "%d", shellpid);
			break;
		case '?':
			snprintf(vbuf, MAX_LINE, "%d", last_result);
			break;
		default:
			// not recognized: the dollar remains literal
			snprintf(vbuf, MAX_LINE, "$%c", **w);
			break;
		}
	}
	// skip past the varname
	while (*(*w)++ != '\0')
		;
	return e;
}

// expand all variables in the word into the obuf.
// Words that consists only of $varname, where varname
// expands to nothing, are removed.
// Return a pointer to the start of the expansion.
char *expand(struct input *s, char *w)
{
	char *e, *r;
	int onlyvars = 0;

	r = s->oscan;
	while (*w != '\0') {
		switch (*w) {
		case '$':
			if (!onlyvars)
				onlyvars = 1;
			if ( (e = getvar(&w)) != NULL )
				while (*e != '\0')
					oput(s, *e++);
			break;
		case '"':
			onlyvars = -1;
			if ( (e = getvar(&w)) != NULL )
				while (*e != '\0')
					oput(s, *e++ | QUOTE);
			break;
		default:
			onlyvars = -1;
			oput(s, *w++);
			break;
		}
	}
	if (onlyvars != 1 || s->oscan != r) {
		oput(s, '\0');
		return r;
	}
	return NULL;
}

// split the word into fields using IFS.  Returns a pointer to the first field
// pointed by *ww, or NULL if there are no more fields.  Advance *ww to the next
// field, or set it to NULL if we have reached the last field. The splitting is in
// place.
char *fields_split(char **ww)
{
	char *e, *d, *w, *t;
	// 'e': pointer to the start of the field
	// 'd': delimiter found
	// 'w': used to scan the input
	// 't': pointer to the IFS value

	if (*ww == NULL)
		return NULL; // no more fields
	w = *ww;
	t = getenv("IFS");
	if (t == NULL)
		t = " \t\n"; // default value of IFS
	// if IFS is "" no splitting is performed
	if (t[0] == '\0' || *w == '\0' /* special case for '' and "" in input */) {
		*ww = NULL;
		return w;
	}
	// skip any IFS whitespace at the beginning of the input
	while (*w != '\0' && (d = strchr(t, *w)) && isspace(*d))
		w++;
	if (*w == '\0') {
		// the input consists only of IFS whitespace: no fields are returned
		*ww = NULL;
		return NULL;
	}
	// we have found a field
	e = w;
	// include all the non-IFS chars
	while (*w != '\0' && !strchr(t, *w))
		w++;
	// terminate the field if we have not reached the end of the input
	// (we overwrite the delimiter in the input string)
	if (*w != '\0')
		*w++ = '\0';
	// skip any adjacent IFS whitespace
	while (*w != '\0' && (d = strchr(t, *w)) && isspace(*d))
		w++;
	// let ww point to the rest of the input, or NULL if we have reached the end
	*ww = (*w == '\0' ? NULL : w);
	return e;
}

void remove_quotes(char *w)
{
	while (*w != '\0')
		*w++ &= ~QUOTE;
}

/* get the file name after a redirection operator */
char *getfname(struct input *s)
{
	char *f;
	/* the file name must be a word */
	if (token(s) != 1) {
		fprintf(stderr, "missing file name\n");
		s->error = 1;
		return NULL;
	}
	f = lasttoken(s);
	if ( (f = expand(s, f)) != NULL)
		remove_quotes(f);
	return f;
}

/* if w is an assignment, return a pointer to the assigned value,
 * NULL otherwise.
 */
char *assignment(char *w)
{
	if (!isalpha(*w) && *w != '_')
		return NULL;
	while (isalnum(*w) || *w == '_')
		w++;
	if (*w != '=')
		return NULL;
	*w = 0;
	return w + 1;
}


int main()
{
	struct input s;
	int t, c_argc;
	char *c_argv[MAX_ARGS + 1];
	char *tok;
	const char *ifile, *ofile;
	int oflags;
	int n_vars, i;
	char *vname[MAX_VARS], *vval[MAX_VARS], *w;
	int stopassign;
	int status;

	shellpid = getpid();

	iinit(&s, 0);
	if (s.interactive) {
		signal(SIGQUIT, SIG_IGN);
		signal(SIGINT, SIG_IGN);
	}
	while (nextcmd(&s) > 0) {
		c_argc = 0;
		ifile = ofile = NULL;
		oflags = O_TRUNC;
		stopassign = 0;
		n_vars = 0;
		while ( (t = token(&s)) != '\n') {
			if (s.error)
				break;
			switch (t) {
			case 1:
				tok = lasttoken(&s);
				if (!stopassign) {
					w = assignment(tok);
					if (w != NULL) {
						if (n_vars >= MAX_VARS) {
							fprintf(stderr, "too many assignments\n");
							s.error = 1;
							continue;
						}
						vname[n_vars] = tok;
						vval[n_vars] = w;
						n_vars++;
						continue;
					}
				}
				stopassign = 1;
				tok = expand(&s, tok);
				while ( (w = fields_split(&tok)) ) {
					if (c_argc >= MAX_ARGS) {
						fprintf(stderr, "too many arguments\n");
						s.error = 1;
						break;
					}
					remove_quotes(w);
					c_argv[c_argc++] = w;
				}
				break;
			case '<':
				if (ifile != NULL) {
					fprintf(stderr, "multiple input redirect\n");
					s.error = 1;
					break;
				}
				if ( (ifile = getfname(&s)) == NULL) {
					fprintf(stderr, "missing file name in redirect\n");
					s.error = 1;
					break;
				}
				break;
			case 'A':
				oflags = O_APPEND;
				/* fallthrough */
			case '>':
				if (ofile != NULL) {
					fprintf(stderr, "multiple output redirect\n");
					s.error = 1;
					break;
				}
				if ( (ofile = getfname(&s)) == NULL) {
					fprintf(stderr, "missing file name in redirect\n");
					s.error = 1;
					break;
				}
				break;
			}
		}
		if (s.error) {
			if (!s.interactive)
				exit(1);
			while (token(&s) != '\n')
					;
			continue;
		}
		c_argv[c_argc] = NULL;
		if (c_argc < 1) {
			c_argv[0] = "";
			for (i = 0; i < n_vars; i++) {
				w = expand(&s, vval[i]);
				if (w == NULL)
					w = "";
				remove_quotes(w);
				setenv(vname[i], w, 1);
			}
		}

		if (strcmp(c_argv[0], "unset") == 0) {
			for (i = 1; i < c_argc; i++)
				unsetenv(c_argv[i]);
			continue;
		}

		if (strcmp(c_argv[0], "cd") == 0) {
			last_result = 0;
			if (c_argc > 2) {
				fprintf(stderr, "usage: cd <dir>\n");
				if (!s.interactive)
					exit(1);
				last_result = 1;
				continue;
			}
			if (c_argc < 2) {
				c_argv[1] = getenv("HOME");
			}
			if (chdir(c_argv[1]) < 0) {
				perror(c_argv[1]);
				last_result = 1;
			}
			continue;
		}

		if (fork()) {
			wait(&status);
			if (WIFEXITED(status)) {
				last_result = WEXITSTATUS(status);
			} else if (WIFSIGNALED(status)) {
				fprintf(stderr, "%s\n", strsignal(WTERMSIG(status)));
				last_result = 0x80 | WTERMSIG(status);
			}
		} else {
			signal(SIGQUIT, SIG_DFL);
			signal(SIGINT, SIG_DFL);

			if (ifile != NULL) {
				close(0);
				if (open(ifile, O_RDONLY) < 0) {
					perror(ifile);
					exit(1);
				}
			}
			if (ofile != NULL) {
				close(1);
				if (open(ofile, O_WRONLY | O_CREAT | oflags, 0666) < 0) {
					perror(ofile);
					exit(1);
				}
			}
			if (c_argc < 1)
				exit(0);
			for (i = 0; i < n_vars; i++) {
				w = expand(&s, vval[i]);
				if (w == NULL)
					w = "";
				remove_quotes(w);
				setenv(vname[i], w, 1);
			}
			execvp(c_argv[0], c_argv);
			perror(c_argv[0]);
			exit(errno == ENOENT ? 127 : 126);
		}
	}
	iclose(&s);
	return last_result;
}
