#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h> /* for perror() */
#include <string.h>
#define _GNU_SOURCE
#include <unistd.h>

#include "input.h"

/* restructuring: data structures for scripts and commands */

#define QUOTE 0x80
#define MAX_ARGS 10
#define MAX_VARS 10

struct script {
	struct input s;
	int argc;
	char *argv[MAX_ARGS + 1];
	int last_result;
	pid_t last_bg_pid;
};

void sinit(struct script *sc, int fd, char *scriptname, char **argv)
{
	iinit(&sc->s, fd);
	sc->argc = 1;
	sc->argv[0] = scriptname;
	if (argv != NULL)
		while (*argv != NULL) {
			sc->argv[sc->argc++] = *argv++;
		}
	sc->argv[sc->argc] = NULL;
	sc->last_result = 0;
	sc->last_bg_pid = -1;
}

int sclose(struct script *sc)
{
	iclose(&sc->s);
	return sc->last_result;
}

struct command {
	int c_argc;
	char *c_argv[MAX_ARGS + 1];
	const char *ifile, *ofile;
	int oflags;
	int n_vars;
	char *vname[MAX_VARS], *vval[MAX_VARS];
	int background;
	pid_t pid;
};

void cinit(struct command *c)
{
	c->c_argc = 0;
	c->ifile = c->ofile = NULL;
	c->oflags = O_TRUNC;
	c->n_vars = 0;
	c->background = 0;
}

pid_t shellpid;

void varname(struct input *s)
{
	if (isalpha(s->ibuf) || s->ibuf == '_') {
		do {
			oput(s, s->ibuf);
			inext(s, 0);
		} while (isalnum(s->ibuf) || s->ibuf == '_');
	} else {
		if (s->ibuf != '\n' && s->ibuf != EOF) {
			oput(s, s->ibuf);
			inext(s, 0);
		}
	}
	// terminate the varname
	oput(s, '\0');
}

void space(struct input *s)
{
	while (s->ibuf == ' ' || s->ibuf == '\t')
		inext(s, 0);
}

void collect_sq(struct input *s)
{
	// skip the initial quote
	inext(s, 1);
	// scan the ibuf until we find the closing single quote,
	// copying and marking all enclosed characters as we go
	while (s->ibuf != '\'') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing quote\n");
			s->error = 1;
			return;
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 1);
	}
	// skip the final quote
	inext(s, 0);
}

void collect_dq(struct input *s)
{
	// skip the initial double quote
	inext(s, 0);
	// scan the ibuf until we find the closing double quote,
	// copying and marking enclosed characters as we go
	while (s->ibuf != '"') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing double quote\n");
			s->error = 1;
			return;
		}
		if (s->ibuf == '$') {
			inext(s, 0);
			if (s->ibuf == '"' || s->ibuf == EOF)
				continue;
			// ugly trick. The '$' may behave in three ways:
			//
			//   1) unquoted: expand the variable without
			//   	         quoting the resulting spaces
			//   2) quoted in dq: expand and quote everything
			//   3) quoted (sq or bs): do nothing
			//
			// but, we only have one spare bit. We use the QUOTE
			// bit to distinguish 1) and 3). For 2) we replace
			// the '$' with '"'. This works, because unQUOTEd
			// double quotes cannot otherwise appear in the obuf.
			oput(s, '"');
			varname(s);
			continue;
		}
		if (s->ibuf == '\\') {
			// a backslash within double quotes removes the
			// special meaning from some characters.
			inext(s, 1);
			if (s->ibuf != '$' && s->ibuf != '"' && s->ibuf != '\\')
				oput(s, '\\' | QUOTE);
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 0);
	}
	// skip the final double quote
	inext(s, 0);
}

int word_delim(int c)
{
	return     c == EOF || isspace(c)
		|| c == '<'  || c == '>'
	        || c == ';'  || c == '&'
		;
}

int word(struct input *s)
{
	newtoken(s);
	while (!word_delim(s->ibuf)) {
		switch (s->ibuf) {
		case '\'':
			collect_sq(s);
			break;
		case '"':
			collect_dq(s);
			break;
		case '$':
			oput(s, '$');
			inext(s, 0);
			varname(s);
			break;
		case '\\':
			// skip the backslash
			inext(s, 1);
			// copy and quote the next char
			oput(s, s->ibuf | QUOTE);
			inext(s, 0);
			break;
		default:
			oput(s, s->ibuf);
			inext(s, 0);
			break;
		}
	}
	// terminate the word
	oput(s, '\0');
	return 1;
}

int redir(struct input *s)
{
	int t;

	t = s->ibuf;
	inext(s, 0);
	if (t == '>') {
		if (s->ibuf == '>') {
			inext(s, 0);
			t = 'A';
		}
	}
	return t;
}

int token(struct input *s)
{
	space(s);
	switch (s->ibuf) {
	case ';':
		inext(s, 0);
		/* fallthrough */
	case '\0':
	case '\n':
		return '\n';
	case '&':
		inext(s, 0);
		return '&';
	case '<': case '>':
		return redir(s);
	default:
		return word(s);
	}
}

// expand the variable pointed to by *w, advancing the *w pointer
// past the terminating '\0' character. Return the variable
// value, or NULL if the variable is not set.
char *getvar(struct script *sc, char **w)
{
	static char vbuf[MAX_LINE];
	char *e;

	// skip past the varname delimiter
	(*w)++;
	// get the value of the variable
	if (isalpha(**w))
		e = getenv(*w);
	else {
		// special variables
		e = vbuf;
		switch (**w) {
		case '$':
			snprintf(vbuf, MAX_LINE, "%d", shellpid);
			break;
		case '?':
			snprintf(vbuf, MAX_LINE, "%d", sc->last_result);
			break;
		case '!':
			if (sc->last_bg_pid < 0)
				vbuf[0] = '\0';
			else
				snprintf(vbuf, MAX_LINE, "%d", sc->last_bg_pid);
			break;
		default:
			// not recognized: the dollar remains literal
			snprintf(vbuf, MAX_LINE, "$%c", **w);
			break;
		}
	}
	// skip past the varname
	while (*(*w)++ != '\0')
		;
	return e;
}

// expand all variables in the word into the obuf.
// Words that consists only of $varname, where varname
// expands to nothing, are removed.
// Return a pointer to the start of the expansion.
char *expand(struct script *sc, char *w)
{
	char *e, *r;
	int onlyvars = 0;
	struct input *s = &sc->s;

	r = s->oscan;
	while (*w != '\0') {
		switch (*w) {
		case '$':
			if (!onlyvars)
				onlyvars = 1;
			if ( (e = getvar(sc, &w)) != NULL )
				while (*e != '\0')
					oput(s, *e++);
			break;
		case '"':
			onlyvars = -1;
			if ( (e = getvar(sc, &w)) != NULL )
				while (*e != '\0')
					oput(s, *e++ | QUOTE);
			break;
		default:
			onlyvars = -1;
			oput(s, *w++);
			break;
		}
	}
	if (onlyvars != 1 || s->oscan != r) {
		oput(s, '\0');
		return r;
	}
	return NULL;
}

// split the word into fields using IFS.  Returns a pointer to the first field
// pointed by *ww, or NULL if there are no more fields.  Advance *ww to the next
// field, or set it to NULL if we have reached the last field. The splitting is in
// place.
char *fields_split(char **ww)
{
	char *e, *d, *w, *t;
	// 'e': pointer to the start of the field
	// 'd': delimiter found
	// 'w': used to scan the input
	// 't': pointer to the IFS value

	if (*ww == NULL)
		return NULL; // no more fields
	w = *ww;
	t = getenv("IFS");
	if (t == NULL)
		t = " \t\n"; // default value of IFS
	// if IFS is "" no splitting is performed
	if (t[0] == '\0' || *w == '\0' /* special case for '' and "" in input */) {
		*ww = NULL;
		return w;
	}
	// skip any IFS whitespace at the beginning of the input
	while (*w != '\0' && (d = strchr(t, *w)) && isspace(*d))
		w++;
	if (*w == '\0') {
		// the input consists only of IFS whitespace: no fields are returned
		*ww = NULL;
		return NULL;
	}
	// we have found a field
	e = w;
	// include all the non-IFS chars
	while (*w != '\0' && !strchr(t, *w))
		w++;
	// terminate the field if we have not reached the end of the input
	// (we overwrite the delimiter in the input string)
	if (*w != '\0')
		*w++ = '\0';
	// skip any adjacent IFS whitespace
	while (*w != '\0' && (d = strchr(t, *w)) && isspace(*d))
		w++;
	// let ww point to the rest of the input, or NULL if we have reached the end
	*ww = (*w == '\0' ? NULL : w);
	return e;
}

void remove_quotes(char *w)
{
	while (*w != '\0')
		*w++ &= ~QUOTE;
}

/* get the file name after a redirection operator */
char *getfname(struct script *sc)
{
	char *f;
	struct input *s = &sc->s;

	/* the file name must be a word */
	if (token(s) != 1) {
		fprintf(stderr, "missing file name\n");
		s->error = 1;
		return NULL;
	}
	f = lasttoken(s);
	if ( (f = expand(sc, f)) != NULL)
		remove_quotes(f);
	return f;
}

/* if w is an assignment, return a pointer to the assigned value,
 * NULL otherwise.
 */
char *assignment(char *w)
{
	if (!isalpha(*w) && *w != '_')
		return NULL;
	while (isalnum(*w) || *w == '_')
		w++;
	if (*w != '=')
		return NULL;
	*w = 0;
	return w + 1;
}

// if 'c' is recognized as internal command, execute
// it and return its status (>= 0), otherwise return -1.
int cinterp_int(struct script *sc, struct command *c)
{
	struct input *s = &sc->s;
	char *w;
	int i;

	if (c->c_argc < 1) {
		for (i = 0; i < c->n_vars; i++) {
			w = expand(sc, c->vval[i]);
			if (w == NULL)
				w = "";
			remove_quotes(w);
			setenv(c->vname[i], w, 1);
		}
		// pretend not recognized if any redirection is needed
		// (the forked process will process the redirections and
		// then exit)
		return (c->ifile != NULL || c->ofile != NULL) ? -1 : 0;
	}
	c->c_argv[c->c_argc] = NULL;

	if (strcmp(c->c_argv[0], "unset") == 0) {
		for (i = 1; i < c->c_argc; i++)
			unsetenv(c->c_argv[i]);
		return 0;
	}

	if (strcmp(c->c_argv[0], "cd") == 0) {
		if (c->c_argc > 2) {
			fprintf(stderr, "syntax error\n");
			if (!s->interactive)
				return 127;
			return 1;
		}
		if (c->c_argc < 2) {
			c->c_argv[1] = getenv("HOME");
		}
		if (chdir(c->c_argv[1]) < 0) {
			perror(c->c_argv[1]);
			return 1;
		}
		return 0;
	}
	return -1; /* not recognized */
}

int cinterp_ext(struct script *sc, struct command *c)
{
	int i;
	char *w;
	struct input *s = &sc->s;

	if (!c->background) {
		signal(SIGQUIT, SIG_DFL);
		signal(SIGINT, SIG_DFL);
	} else {
		if (c->ifile == NULL && s->interactive)
			c->ifile = "/dev/null";
	}

	if (c->ifile != NULL) {
		close(0);
		if (open(c->ifile, O_RDONLY) < 0) {
			perror(c->ifile);
			exit(1);
		}
	}
	if (c->ofile != NULL) {
		close(1);
		if (open(c->ofile, O_WRONLY | O_CREAT | c->oflags, 0666) < 0) {
			perror(c->ofile);
			exit(1);
		}
	}
	if (c->c_argc < 1)
		return 0;
	for (i = 0; i < c->n_vars; i++) {
		w = expand(sc, c->vval[i]);
		if (w == NULL)
			w = "";
		remove_quotes(w);
		setenv(c->vname[i], w, 1);
	}
	execvp(c->c_argv[0], c->c_argv);
	perror(c->c_argv[0]);
	return (errno == ENOENT ? 127 : 126);
}

int sinterp(struct script *sc)
{
	char *tok;
	int t;
	char *w;
	int stopassign;
	struct command cc;
	struct input *s = &sc->s;
	int status;
	struct command *c;

	while (nextcmd(s) > 0) {
		c = &cc;
		cinit(c);
		stopassign = 0;
		while ( (t = token(s)) != '\n') {
			if (s->error)
				break;
			switch (t) {
			case 1:
				tok = lasttoken(s);
				if (!stopassign) {
					w = assignment(tok);
					if (w != NULL) {
						if (c->n_vars >= MAX_VARS) {
							fprintf(stderr, "too many assignments\n");
							s->error = 1;
							continue;
						}
						c->vname[c->n_vars] = tok;
						c->vval[c->n_vars] = w;
						c->n_vars++;
						continue;
					}
				}
				stopassign = 1;
				tok = expand(sc, tok);
				while ( (w = fields_split(&tok)) ) {
					if (c->c_argc >= MAX_ARGS) {
						fprintf(stderr, "too many arguments\n");
						s->error = 1;
						break;
					}
					remove_quotes(w);
					c->c_argv[c->c_argc++] = w;
				}
				break;
			case '<':
				if (c->ifile != NULL) {
					fprintf(stderr, "multiple input redirect\n");
					s->error = 1;
					break;
				}
				if ( (c->ifile = getfname(sc)) == NULL) {
					fprintf(stderr, "missing filename in redirect\n");
					s->error = 1;
					break;
				}
				break;
			case 'A':
				c->oflags = O_APPEND;
				/* fallthrough */
			case '>':
				if (c->ofile != NULL) {
					fprintf(stderr, "multiple output redirect\n");
					s->error = 1;
					break;
				}
				if ( (c->ofile = getfname(sc)) == NULL) {
					fprintf(stderr, "missing filename in redirect\n");
					s->error = 1;
					break;
				}
				break;
			case '&':
				c->background = 1;
				goto run;
			}
		}
	run:
		if (s->error) {
			if (!s->interactive)
				exit(1);
			while (token(s) != '\n')
					;
			continue;
		}

		sc->last_result = cinterp_int(sc, c);
		if (sc->last_result >= 0)
			continue;
		if ( (c->pid = fork()) ) {
			if (c->background) {
				sc->last_bg_pid = c->pid;
				sc->last_result = 0;
				continue;
			}
			while (wait(&status) != c->pid)
				;
			if (WIFEXITED(status)) {
				sc->last_result = WEXITSTATUS(status);
			} else if (WIFSIGNALED(status)) {
				fprintf(stderr, "%s\n", strsignal(WTERMSIG(status)));
				sc->last_result = 0x80 | WTERMSIG(status);
			}
		} else {
			exit(cinterp_ext(sc, c));
		}
	}
	return sc->last_result;
}

int main(int argc, char *argv[])
{
	struct script sc;

	shellpid = getpid();

	sinit(&sc, 0, argv[0], argv + 1);
	if (sc.s.interactive) {
		signal(SIGQUIT, SIG_IGN);
		signal(SIGINT, SIG_IGN);
	}
	sinterp(&sc);
	return sclose(&sc);
}
