#include <sys/wait.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <setjmp.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h> /* for perror() */
#include <string.h>
#define _GNU_SOURCE
#include <unistd.h>

#include "input.h"

/* subshells and command substitution */

#define QUOTE 0x80
#define MAX_ARGS 100
#define MAX_VARS 10
#define MAX_PIPES 10

#define unquote(p__) (*(unsigned char *)(p__) & ~QUOTE)

struct script {
	struct input *s;
	int argc;
	char *argv[MAX_ARGS + 1];
	int last_result;
	pid_t last_bg_pid;
	char args[MAX_LINE];
};

void sinit(struct script *sc, struct input *s, char *scriptname, char **argv)
{
	sc->s = s;
	sc->argc = 1;
	sc->argv[0] = scriptname;
	if (argv != NULL)
		while (*argv != NULL) {
			sc->argv[sc->argc++] = *argv++;
		}
	sc->argv[sc->argc] = NULL;
	sc->last_result = 0;
	sc->last_bg_pid = -1;
}

int sclose(struct script *sc)
{
	iclose(sc->s);
	return sc->last_result;
}

struct command {
	int c_argc;
	char *c_argv[MAX_ARGS + 1];
	const char *ifile, *ofile;
	int oflags;
	int n_vars;
	char *vname[MAX_VARS], *vval[MAX_VARS];
	int background;
	pid_t pid;
};

void cinit(struct command *c)
{
	c->c_argc = 0;
	c->ifile = c->ofile = NULL;
	c->oflags = O_TRUNC;
	c->n_vars = 0;
	c->background = 0;
}

pid_t shellpid;
void copy_to(struct input *s, char q);

void varname(struct input *s)
{
	if (isalpha(s->ibuf) || s->ibuf == '_') {
		do {
			oput(s, s->ibuf);
			inext(s, 0);
		} while (isalnum(s->ibuf) || s->ibuf == '_');
	} else if (s->ibuf == '(') {
		oput(s, s->ibuf);
		inext(s, 0);
		copy_to(s, ')');
		oput(s, '\n');
		inext(s, 0);
	} else {
		if (s->ibuf != '\n' && s->ibuf != EOF) {
			oput(s, s->ibuf);
			inext(s, 0);
		}
	}
	// terminate the varname
	oput(s, '\0');
}

void comment(struct input *s)
{
	do {
		inext(s, 0);
	} while (s->ibuf != '\n');
}

void space(struct input *s)
{
	for (;;) {
		switch (s->ibuf) {
		case ' ': case '\t':
			break;
		case '#':
			if (!s->interactive) {
				comment(s);
				break;
			}
			/* fallthrough */
		default:
			return;
		}
		inext(s, 0);
	}
}

void space_or_newline(struct input *s)
{
	space(s);
	while (s->ibuf == '\n' && s->ibuf != EOF) {
		inext(s, 0);
		space(s);
	}
}

void collect_sq(struct input *s)
{
	// skip the initial quote
	inext(s, 1);
	// scan the ibuf until we find the closing single quote,
	// copying and marking all enclosed characters as we go
	while (s->ibuf != '\'') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing quote\n");
			s->error = 1;
			return;
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 1);
	}
	// skip the final quote
	inext(s, 0);
}

void collect_dq(struct input *s)
{
	// skip the initial double quote
	inext(s, 0);
	// scan the ibuf until we find the closing double quote,
	// copying and marking enclosed characters as we go
	while (s->ibuf != '"') {
		if (s->ibuf == EOF) {
			fprintf(stderr, "missing closing double quote\n");
			s->error = 1;
			return;
		}
		if (s->ibuf == '$') {
			inext(s, 0);
			if (s->ibuf == '"' || s->ibuf == EOF)
				continue;
			// ugly trick. The '$' may behave in three ways:
			//
			//   1) unquoted: expand the variable without
			//   	         quoting the resulting spaces
			//   2) quoted in dq: expand and quote everything
			//   3) quoted (sq or bs): do nothing
			//
			// but, we only have one spare bit. We use the QUOTE
			// bit to distinguish 1) and 3). For 2) we replace
			// the '$' with '"'. This works, because unQUOTEd
			// double quotes cannot otherwise appear in the obuf.
			oput(s, '"');
			varname(s);
			continue;
		}
		if (s->ibuf == '\\') {
			// a backslash within double quotes removes the
			// special meaning from some characters.
			inext(s, 1);
			if (s->ibuf != '$' && s->ibuf != '"' && s->ibuf != '\\')
				oput(s, '\\' | QUOTE);
		}
		oput(s, s->ibuf | QUOTE);
		inext(s, 0);
	}
	// skip the final double quote
	inext(s, 0);
}

int word_delim(struct input *s, int c)
{
	return     c == EOF || isspace(c)
		|| c == '<'  || c == '>'
	        || c == ';'  || c == '&'
	        || (!s->interactive && c == '#')
		|| c == '|'
		|| c == '(' || c == ')'
		;
}

int word(struct input *s)
{
	newtoken(s);
	while (!word_delim(s, s->ibuf)) {
		switch (s->ibuf) {
		case '\'':
			collect_sq(s);
			break;
		case '"':
			collect_dq(s);
			break;
		case '$':
			oput(s, '$');
			inext(s, 0);
			varname(s);
			break;
		case '\\':
			// skip the backslash
			inext(s, 1);
			// copy and quote the next char
			oput(s, s->ibuf | QUOTE);
			inext(s, 0);
			break;
		default:
			oput(s, s->ibuf);
			inext(s, 0);
			break;
		}
	}
	// terminate the word
	oput(s, '\0');
	return 1;
}

int redir(struct input *s)
{
	int t;

	t = s->ibuf;
	inext(s, 0);
	if (t == '>') {
		if (s->ibuf == '>') {
			inext(s, 0);
			t = 'A';
		}
	}
	return t;
}

void copy_to(struct input *s, char q)
{
	while (!s->error && s->ibuf != q) {
		if (s->ibuf == EOF) {
			fprintf(stderr, "syntax error\n");
			s->error = 1;
			break;
		}
		switch (s->ibuf) {
		case '\\':
			if (q != '\'') {
				oput(s, '\\');
				inext(s, 1);
			}
			break;
		case '"':
			if (q == '\'')
				break;
			oput(s, '"');
			inext(s, 1);
			copy_to(s, '"');
			break;
		case '\'':
			if (q == '"')
				break;
			oput(s, '\'');
			inext(s, 1);
			copy_to(s, '\'');
			break;
		case '$':
			if (q == '\'')
				break;
			oput(s, '$');
			inext(s, 0);
			if (s->ibuf != '(')
				continue;
			oput(s, '(');
			inext(s, 1);
			copy_to(s, ')');
			break;
		case '(':
			if (q == '\'' || q == '"')
				break;
			oput(s, '(');
			inext(s, 1);
			copy_to(s, ')');
			break;
		}
		oput(s, s->ibuf);
		inext(s, q == '\'');
	}
}

int collect_subsh(struct input *s)
{
	newtoken(s);
	inext(s, 0);
	copy_to(s, ')');
	inext(s, 0);
	oput(s, '\0');
	return '(';
}

int token(struct input *s)
{
	space(s);
	switch (s->ibuf) {
	case '(':
		return collect_subsh(s);
	case ')':
		return ')';
	case ';':
		inext(s, 0);
		/* fallthrough */
	case EOF:
	case '\n':
		return '\n';
	case '&':
		inext(s, 0);
		if (s->ibuf == '&') {
			inext(s, 0);
			space_or_newline(s);
			return 'a';
		}
		return '&';
	case '|':
		inext(s, 0);
		if (s->ibuf == '|') {
			inext(s, 0);
			space_or_newline(s);
			return 'o';
		}
		space_or_newline(s);
		return '|';
	case '<': case '>':
		return redir(s);
	default:
		return word(s);
	}
}

size_t join_args(char *buf, int m, int argc, char *argv[])
{
	int i, t, n;
	char *v;

	v = buf;
	t = MAX_LINE - 1;
	for (i = 1; i < argc; i++) {
		if (i > 1 && t > 1) {
			*v++ = ' ';
			t--;
		}
		if (t <= 0)
			break;
		n = snprintf(v, t, "%s", argv[i]);
		if (n >= t)
			break;
		t -= n;
		while (n--)
			*v++ |= m;
	}
	*v = '\0';
	return v - buf;
}

int sinterp(struct script *sc);
void command_substition(struct script *sc, char *e, const char *c)
{
	int p[2];
	int n, r;
	struct input s;
	struct script sc1;

	if (pipe(p) < 0)
		return;

	switch (fork()) {
	case -1:
		return;
	case 0:
		close(0);
		open("/dev/null", O_RDONLY);
		close(1);
		dup(p[1]);
		close(p[0]);
		close(p[1]);
		iinit_mem(&s, c);
		sinit(&sc1, &s, sc->argv[0], sc->argv + 1);
		sinterp(&sc1);
		exit(sclose(&sc1));
		break;
	default:
		close(p[1]);
		n = 0;
		while (n < MAX_LINE - 1 && (r = read(p[0], e + n, MAX_LINE - n - 1)) > 0) {
			n += r;
		}
		e[n] = '\0';
		while (n > 0 && e[n - 1] == '\n')
			e[--n] = '\0';
		close(p[0]);
		wait(NULL);
		break;
	}
}

// expand the variable pointed to by *w, advancing the *w pointer
// past the terminating '\0' character. Return the variable
// value, or NULL if the variable is not set.
char *getvar(struct script *sc, char **w)
{
	static char vbuf[MAX_LINE];
	char q, *e;
	int i, m;

	// skip past the varname delimiter
	q = *(*w)++;
	m = 0;
	// get the value of the variable
	if (isalpha(**w))
		e = getenv(*w);
	else {
		// special variables
		e = vbuf;
		switch (**w) {
		case '$':
			snprintf(vbuf, MAX_LINE, "%d", shellpid);
			break;
		case '?':
			snprintf(vbuf, MAX_LINE, "%d", sc->last_result);
			break;
		case '!':
			if (sc->last_bg_pid < 0)
				vbuf[0] = '\0';
			else
				snprintf(vbuf, MAX_LINE, "%d", sc->last_bg_pid);
			break;
		case '#':
			snprintf(vbuf, MAX_LINE, "%d", sc->argc - 1);
			break;
		case '@':
			if (q == '"')
				m = QUOTE;
			/* fallthrough */
		case '*':
			join_args(vbuf, m, sc->argc, sc->argv);
			break;
		case '0':
			e = sc->argv[0];
			break;
		case '(':
			command_substition(sc, e, (*w)+1);
			break;
		default:
			if (**w >= '1' && **w <= '9') {
				i = **w - '0';
				if (i < sc->argc) {
					e = sc->argv[i];
				} else {
					vbuf[0] = '\0';
				}
				break;
			}
			// not recognized: the dollar remains literal
			snprintf(vbuf, MAX_LINE, "$%c", **w);
			break;
		}
	}
	// skip past the varname
	while (*(*w)++ != '\0')
		;
	return e;
}

// expand all variables in the word into the obuf.
// Words that consists only of $varname, where varname
// expands to nothing, are removed.
// Return a pointer to the start of the expansion.
char *expand(struct script *sc, char *w)
{
	char *e, *r;
	int onlyvars = 0;
	struct input *s = sc->s;

	r = s->oscan;
	while (*w != '\0') {
		switch (*w) {
		case '$':
		unquoted:
			if (!onlyvars)
				onlyvars = 1;
			if ( (e = getvar(sc, &w)) != NULL )
				while (*e != '\0')
					oput(s, *e++);
			break;
		case '"':
			if (*(w + 1) == '@')
				goto unquoted;
			onlyvars = -1;
			if ( (e = getvar(sc, &w)) != NULL )
				while (*e != '\0')
					oput(s, *e++ | QUOTE);
			break;
		default:
			onlyvars = -1;
			oput(s, *w++);
			break;
		}
	}
	if (onlyvars != 1 || s->oscan != r) {
		oput(s, '\0');
		return r;
	}
	return NULL;
}

// split the word into fields using IFS.  Returns a pointer to the first field
// pointed by *ww, or NULL if there are no more fields.  Advance *ww to the next
// field, or set it to NULL if we have reached the last field. The splitting is in
// place.
char *fields_split(char **ww)
{
	char *e, *d, *w, *t;
	// 'e': pointer to the start of the field
	// 'd': delimiter found
	// 'w': used to scan the input
	// 't': pointer to the IFS value

	if (*ww == NULL)
		return NULL; // no more fields
	w = *ww;
	t = getenv("IFS");
	if (t == NULL)
		t = " \t\n"; // default value of IFS
	// if IFS is "" no splitting is performed
	if (t[0] == '\0' || *w == '\0' /* special case for '' and "" in input */) {
		*ww = NULL;
		return w;
	}
	// skip any IFS whitespace at the beginning of the input
	while (*w != '\0' && (d = strchr(t, *w)) && isspace(*d))
		w++;
	if (*w == '\0') {
		// the input consists only of IFS whitespace: no fields are returned
		*ww = NULL;
		return NULL;
	}
	// we have found a field
	e = w;
	// include all the non-IFS chars
	while (*w != '\0' && !strchr(t, *w))
		w++;
	// terminate the field if we have not reached the end of the input
	// (we overwrite the delimiter in the input string)
	if (*w != '\0')
		*w++ = '\0';
	// skip any adjacent IFS whitespace
	while (*w != '\0' && (d = strchr(t, *w)) && isspace(*d))
		w++;
	// let ww point to the rest of the input, or NULL if we have reached the end
	*ww = (*w == '\0' ? NULL : w);
	return e;
}

int match1(char* p, char* n)
{
	if (!*p)
		return !*n;

	if (*p == '*') {
		if (*n && match1(p, n + 1))
			return 1;
		return match1(p + 1, n);
	}

	if (!*n)
		return 0;

	if (*p == '?' || unquote(p) == *n)
		return match1(p + 1, n + 1);

	return 0;
}

int match(char* p, char* n)
{
	if (*n == '.') {
		if (unquote(p) != '.')
			return 0;
		else {
			p++;
			n++;
		}
	}
	return match1(p, n);
}

int ispattern(char *p)
{
	for (; *p; p++)
		if (*p == '?' || *p == '*')
			return 1;
	return 0;
}

int glob(struct input *s, DIR *d, char *p, char *w, char **argv, int *argc, int max)
{
	DIR *sd;
	struct dirent* e;
	size_t lp, l, i;
	char *start, *end, *save_oscan, old, *nw;
	int n, m;

	lp = strlen(p);

	for (end = w; *end != '\0' && unquote(end) != '/'; end++)
		;

	old = *end;
	*end = '\0';

	n = 0;
	while ( (e = readdir(d)) ) {
		if (match(w, e->d_name)) {
			l  = strlen(e->d_name);
			if ((n ? lp : 0) + l >= ospace(s)) {
				fprintf(stderr, "line too long\n");
				s->error = 1;
				break;
			}
			save_oscan = s->oscan;
			if (n) {
				start = s->oscan;
				for (i = 0; i < lp; i++)
					*s->oscan++ = p[i];
			} else {
				start = p;
			}
			strcpy(s->oscan, e->d_name);
			s->oscan += l;
			if (old == '\0') {
				if (*argc >= max) {
					fprintf(stderr, "too many pathnames\n");
					s->error = 1;
					break;
				}
				n++;
				s->oscan++;
				argv[(*argc)++] = start;
				continue;
			}
			m = 0;
			if ( (sd = opendir(start)) != NULL ) {
				oput(s, '/');
				if (s->error)
					break;
				*s->oscan = '\0';
				for (nw = end + 1; unquote(nw) == '/'; nw++)
					;
				m = glob(s, sd, start, nw, argv, argc, max);
			}
			if (!m)
				s->oscan = save_oscan;
			n += m;
		}
	}
	*end = old;
	closedir(d);
	return n;
}

int pathnames(struct input *s, char *w, char **argv, int *argc, int max)
{
	DIR *d;
	char *p, *ww;
	int oargc;

	if (!ispattern(w))
		return 0;

	p = s->oscan;
	if (unquote(w) == '/')
		oput(s, '/');
	if (s->error)
		return -1;
	*s->oscan = '\0';
	if ( (d = opendir(unquote(w) == '/' ? "/" : ".")) == NULL) {
		perror("glob");
		return -1;
	}
	ww = w;
	while (unquote(w) == '/')
		w++;

	oargc = *argc;
	glob(s, d, p, w, argv, argc, max);
	if (oargc == *argc)
		argv[(*argc)++] = ww;
	return 1;
}

void remove_quotes(char *w)
{
	while (*w != '\0')
		*w++ &= ~QUOTE;
}

/* get the file name after a redirection operator */
char *getfname(struct script *sc)
{
	char *f, *w;
	int n;
	struct input *s = sc->s;

	/* the file name must be a word */
	if (token(s) != 1) {
		fprintf(stderr, "missing file name\n");
		s->error = 1;
		return NULL;
	}
	f = lasttoken(s);
	w = NULL;
	if ( (f = expand(sc, f)) != NULL) {
		n = 0;
		switch (pathnames(s, f, &w, &n, 1)) {
		case 0:
			w = f;
			break;
		case 1:
			break;
		default:
			return NULL;
		}
		remove_quotes(w);
	}
	return w;
}

/* if w is an assignment, return a pointer to the assigned value,
 * NULL otherwise.
 */
char *assignment(char *w)
{
	if (!isalpha(*w) && *w != '_')
		return NULL;
	while (isalnum(*w) || *w == '_')
		w++;
	if (*w != '=')
		return NULL;
	*w = 0;
	return w + 1;
}

jmp_buf dot_jb;
void int_handler(int signo)
{
	longjmp(dot_jb, 1);
}

int cinterp_ext(struct script *sc, struct command *c);
char *nextpath(const char **paths, const char *file);
// if 'c' is recognized as internal command, execute
// it and return its status (>= 0), otherwise return -1.
int cinterp_int(struct script *sc, struct command *c)
{
	struct input *s = sc->s;
	char *w;
	int i, n;
	char dbuf[MAX_LINE];
	struct script sd;
	int fd;
	struct input isd;

	if (c->c_argc < 1) {
		for (i = 0; i < c->n_vars; i++) {
			w = expand(sc, c->vval[i]);
			if (w == NULL)
				w = "";
			remove_quotes(w);
			setenv(c->vname[i], w, 1);
		}
		// pretend not recognized if any redirection is needed
		// (the forked process will process the redirections and
		// then exit)
		return (c->ifile != NULL || c->ofile != NULL) ? -1 : 0;
	}
	c->c_argv[c->c_argc] = NULL;

	if (strcmp(c->c_argv[0], "unset") == 0) {
		for (i = 1; i < c->c_argc; i++)
			unsetenv(c->c_argv[i]);
		return 0;
	}

	if (strcmp(c->c_argv[0], "cd") == 0) {
		if (c->c_argc > 2) {
			fprintf(stderr, "syntax error\n");
			if (!s->interactive)
				return 127;
			return 1;
		}
		if (c->c_argc < 2) {
			c->c_argv[1] = getenv("HOME");
			if (c->c_argv[1] == NULL)
				c->c_argv[1] = "/";
		}
		if (chdir(c->c_argv[1]) < 0) {
			perror(c->c_argv[1]);
			return 1;
		}
		return 0;
	}

	if (strcmp(c->c_argv[0], "shift") == 0) {
		if (c->c_argc > 2) {
			fprintf(stderr, "syntax error\n");
			return 1;
		}
		n = 1;
		if (c->c_argc > 1) {
			n = atoi(c->c_argv[1]);
			if (n <= 0) {
				fprintf(stderr, "syntax error\n");
				return 1;
			}
		}
		if (sc->argc - 1 < n) {
			fprintf(stderr, "cannot shift\n");
			return 1;
		}
		for (i = n + 1; i <= sc->argc; i++)
			sc->argv[i - n] = sc->argv[i];
		sc->argc -= n;
		return 0;
	}

	if (strcmp(c->c_argv[0], ".") == 0) {
		const char *p, *ss;

		if (c->c_argc < 2) {
			fprintf(stderr, "syntax error\n");
			return 1;
		}
		p = getenv("PATH");
		if (p == NULL || *p == '\0' || strchr(c->c_argv[1], '/')) {
			fd = open(c->c_argv[1], O_RDONLY);
		} else {
			fd = -1;
			while ( (ss = nextpath(&p, c->c_argv[1])) != NULL ) {
				if ( (fd = open(ss, O_RDONLY)) >= 0)
					break;
			}
		}
		if (fd < 0) {
			perror(c->c_argv[1]);
			return 1;
		}
		iinit(&isd, fd);
		if (s->interactive) {
			if (setjmp(dot_jb)) {
				signal(SIGINT, SIG_IGN);
				return 0x80 | SIGINT;
			}
			signal(SIGINT, int_handler);
		}
		sinit(&sd, &isd, c->c_argv[1], c->c_argv + 2);
		sinterp(&sd);
		if (s->interactive)
			signal(SIGINT, SIG_IGN);
		return sclose(&sd);
	}

	if (strcmp(c->c_argv[0], "eval") == 0) {
		join_args(dbuf, 0, c->c_argc, c->c_argv);
		iinit_mem(&isd, dbuf);
		sinit(&sd, &isd, sc->argv[0], NULL);
		sinterp(&sd);
		return sclose(&sd);
	}

	if (strcmp(c->c_argv[0], "exec") == 0) {
		for (i = 1; i <= c->c_argc; i++)
			c->c_argv[i - 1] = c->c_argv[i];
		c->c_argc--;
		return cinterp_ext(sc, c);
	}

	if (strcmp(c->c_argv[0], "umask") == 0) {
		mode_t m;

		if (c->c_argc > 2) {
			fprintf(stderr, "syntax error\n");
			return 1;
		}
		m = umask(0);
		if (c->c_argc == 1) {
			printf("0%03o\n", m);
		} else {
			m = strtol(c->c_argv[1], NULL, 8);
		}
		umask(m);
		return 0;
	}

	if (strcmp(c->c_argv[0], ":") == 0) {
		return 0;
	}

	if (strcmp(c->c_argv[0], "exit") == 0) {
		int rv;

		if (c->c_argc > 2) {
			fprintf(stderr, "syntax error\n");
			return 1;
		}
		rv = sc->last_result;
		if (c->c_argc == 2)
			rv = atoi(c->c_argv[1]);
		exit(rv);
	}

	if (strcmp(c->c_argv[0], "set") == 0) {
		char *a, *b;
		int rv;

		rv = 0;
		sc->argc = 1;
		b = sc->args;
		for (i = 1; i < c->c_argc; i++) {
			a = c->c_argv[i];
			sc->argv[sc->argc++] = b;
			while (b - sc->args < MAX_LINE - 1 && *a != '\0')
				*b++ = *a++;
			*b++ = '\0';
		}
		if (b - sc->args >= MAX_LINE) {
			fprintf(stderr, "line too long\n");
			sc->argc = 1;
			rv = 1;
		}
		sc->argv[sc->argc] = NULL;
		return rv;
	}

	return -1; /* not recognized */
}

int cinterp_ext(struct script *sc, struct command *c)
{
	int i;
	char *w;
	struct input *s = sc->s;

	if (!c->background) {
		signal(SIGQUIT, SIG_DFL);
		signal(SIGINT, SIG_DFL);
	} else {
		if (c->ifile == NULL && s->interactive)
			c->ifile = "/dev/null";
	}

	if (c->ifile != NULL) {
		close(0);
		if (open(c->ifile, O_RDONLY) < 0) {
			perror(c->ifile);
			return 1;
		}
	}
	if (c->ofile != NULL) {
		close(1);
		if (open(c->ofile, O_WRONLY | O_CREAT | c->oflags, 0666) < 0) {
			perror(c->ofile);
			return 1;
		}
	}
	if (c->c_argc < 1)
		return 0;
	for (i = 0; i < c->n_vars; i++) {
		w = expand(sc, c->vval[i]);
		if (w == NULL)
			w = "";
		remove_quotes(w);
		setenv(c->vname[i], w, 1);
	}
	if (c->c_argc < 1)
		return 0;
	execvp(c->c_argv[0], c->c_argv);
	perror(c->c_argv[0]);
	return (errno == ENOENT ? 127 : 126);
}

int sinterp(struct script *sc)
{
	char *tok;
	int t;
	char *w;
	int stopassign;
	struct command cc[MAX_PIPES + 1];
	struct input *s = sc->s;
	int status;
	struct command *c;
	int npipe;
	int p[2], lastpipe, i;
	int operator;
	int negated;

	while (nextcmd(s) > 0) {
		c = &cc[0];
		npipe = 0;
		cinit(c);
		stopassign = 0;
		operator = 0;
		negated = -1;
		while ( (t = token(s)) != '\n') {
			if (s->error)
				break;
			switch (t) {
			case 1:
				tok = lasttoken(s);
				if (!strcmp(tok, "!")) {
					if (negated >= 0) {
						fprintf(stderr, "syntax error\n");
						s->error = 1;
					} else {
						negated = 1;
					}
					continue;
				}
				if (negated < 0)
					negated = 0;
				if (!stopassign) {
					w = assignment(tok);
					if (w != NULL) {
						if (c->n_vars >= MAX_VARS) {
							fprintf(stderr, "too many assignments\n");
							s->error = 1;
							continue;
						}
						c->vname[c->n_vars] = tok;
						c->vval[c->n_vars] = w;
						c->n_vars++;
						continue;
					}
				}
				stopassign = 1;
				tok = expand(sc, tok);
				while ( (w = fields_split(&tok)) ) {
					if (c->c_argc >= MAX_ARGS) {
						fprintf(stderr, "too many arguments\n");
						s->error = 1;
						break;
					}
					if (!pathnames(s, w, c->c_argv, &c->c_argc, MAX_ARGS)) {
						remove_quotes(w);
						c->c_argv[c->c_argc++] = w;
					}
				}
				break;
			case '<':
				if (c->ifile != NULL) {
					fprintf(stderr, "multiple input redirect\n");
					s->error = 1;
					break;
				}
				if ( (c->ifile = getfname(sc)) == NULL) {
					fprintf(stderr, "missing filename in redirect\n");
					s->error = 1;
					break;
				}
				break;
			case 'A':
				c->oflags = O_APPEND;
				/* fallthrough */
			case '>':
				if (c->ofile != NULL) {
					fprintf(stderr, "multiple output redirect\n");
					s->error = 1;
					break;
				}
				if ( (c->ofile = getfname(sc)) == NULL) {
					fprintf(stderr, "missing filename in redirect\n");
					s->error = 1;
					break;
				}
				break;
			case '&':
				c->background = 1;
				goto run;
			case '|':
				if (npipe >= MAX_PIPES) {
					fprintf(stderr, "too many pipes\n");
					s->error = 1;
					break;
				}
				npipe++;
				c++;
				cinit(c);
				stopassign = 0;
				break;
			case 'a': case 'o':
				operator = t;
				goto run;
			case '(':
				if (c->n_vars > 0 || c->c_argc > 0) {
					fprintf(stderr, "syntax error\n");
					s->error = 1;
					break;
				}
				c->c_argv[0] = "/bin/sh";
				c->c_argv[1] = "-c";
				c->c_argv[2] = lasttoken(s);
				c->c_argc = 3;
				break;
			}
		}
	run:
		if ((npipe || operator) && s->ibuf == EOF) {
			fprintf(stderr, "unexpected eof\n");
			s->error = 1;
		}

		if (s->error) {
			if (!s->interactive)
				exit(1);
			while (token(s) != '\n')
					;
			continue;
		}

		if (!npipe) {
			sc->last_result = cinterp_int(sc, c);
			if (sc->last_result >= 0) {
				if (negated > 0)
					sc->last_result = !sc->last_result;
				continue;
			}
		}
		if ( (c->pid = fork()) ) {
			if (c->background) {
				sc->last_bg_pid = c->pid;
				sc->last_result = 0;
				continue;
			}
			while (wait(&status) != c->pid)
				;
			if (WIFEXITED(status)) {
				sc->last_result = WEXITSTATUS(status);
			} else if (WIFSIGNALED(status)) {
				fprintf(stderr, "%s\n", strsignal(WTERMSIG(status)));
				sc->last_result = 0x80 | WTERMSIG(status);
			}
		} else {
			lastpipe = -1;
			sc->last_result = 0;
			if (npipe) {
				cc[0].background = cc[npipe].background;
				cc[npipe].background = 0;
			}
			for (i = 0, c = &cc[0]; i < npipe; i++, c++) {
				if (pipe(p) < 0) {
					perror("pipe");
					sc->last_result = 1;
					break;
				}
				if ( (c->pid = fork()) ) {
					if (lastpipe > 0)
						close(lastpipe);
					lastpipe = p[0];
					close(p[1]);
				} else {
					close(p[0]);
					close(1);
					if (dup(p[1]) < 0) {
						perror("dup");
						exit(1);
					}
					close(p[1]);
					if (lastpipe > 0) {
						close(0);
						if (dup(lastpipe) < 0) {
							perror("dup");
							exit(1);
						}
						close(lastpipe);
					}
					if (cinterp_int(sc, c) < 0)
						cinterp_ext(sc, c);
					exit(0);
				}
			}
			if (!sc->last_result) {
				if (lastpipe > 0) {
					close(0);
					if (dup(lastpipe) < 0) {
						perror("dup");
						exit(1);
					}
					close(lastpipe);
				}
				sc->last_result = npipe ? cinterp_int(sc, c) : -1;
				if (sc->last_result < 0)
					sc->last_result = cinterp_ext(sc, c);
			}
			exit(sc->last_result);
		}
		if (negated > 0)
			sc->last_result = !sc->last_result;
		if (!operator)
			continue;
		if ((operator == 'a' &&  sc->last_result) ||
		    (operator == 'o' && !sc->last_result)) {
			while (t = token(s), t != '\n' && t != 'a' && t != 'o')
				;
		}
	}
	return sc->last_result;
}

int main(int argc, char *argv[])
{
	int i, fd, fdp;
	char *o, *cmd, *name;
	struct input s, sp;
	struct script sc, scp;
	char pp[MAX_LINE];

	if (argc < 1)
		return 1;

	name = argv[0];
	cmd = NULL;
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			break;

		for (o = &argv[i][1]; *o != '\0'; o++) {
			switch (*o) {
			case 'c':
				if (i + 1 < argc) {
					i++;
					cmd = argv[i];
				} else {
					fprintf(stderr, "missing command after -c\n");
					exit(1);
				}
				break;
			default:
				fprintf(stderr, "unknown option: %c\n", *o);
				exit(1);
			}
		}
	}
	argc -= i - 1;
	argv += i - 1;

	if (cmd != NULL) {
		iinit_mem(&s, cmd);
		fd = -1;
	} else if (argc > 1) {
		name = argv[1];
		if ( (fd = open(name, O_RDONLY)) < 0) {
			perror(name);
			exit(errno == ENOENT ? 127: 126);
		}
		argc--;
		argv++;
	} else {
		fd = 0;
	}
	if (fd >= 0)
		iinit(&s, fd);

	shellpid = getpid();

	sinit(&sc, &s, name, argv + 1);
	if (sc.s->interactive) {
		if (name[0] == '-') {
			snprintf(pp, MAX_LINE, "%s/.profile", getenv("HOME"));
			fdp = open(pp, O_RDONLY);
			if (fdp >= 0) {
				iinit(&sp, fdp);
				sinit(&scp, &sp, pp, NULL);
				sinterp(&scp);
				sclose(&scp);
			}
		}
		signal(SIGQUIT, SIG_IGN);
		signal(SIGINT, SIG_IGN);
	}
	sinterp(&sc);
	return sclose(&sc);
}
